package cc.smtweb.system.bpm.web.design.form;

import cc.smtweb.framework.core.SwException;
import cc.smtweb.framework.core.cache.AbstractCache;
import cc.smtweb.framework.core.cache.CacheManager;
import cc.smtweb.framework.core.common.SwEnum;
import cc.smtweb.framework.core.util.JsonUtil;
import cc.smtweb.framework.core.util.MapUtil;
import cc.smtweb.system.bpm.web.design.form.define.PageDataSet;
import cc.smtweb.system.bpm.web.design.form.define.PageDatasetField;
import cc.smtweb.system.bpm.web.design.form.define.PageInfo;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ser.FilterProvider;
import com.fasterxml.jackson.databind.ser.PropertyWriter;
import com.fasterxml.jackson.databind.ser.impl.SimpleBeanPropertyFilter;
import com.fasterxml.jackson.databind.ser.impl.SimpleFilterProvider;
import org.apache.commons.lang3.StringUtils;

import java.util.List;
import java.util.Map;

/**
 * Created by Akmm at 2022/4/20 18:45
 * 辅助类
 */
public class ModelFormHelper {
    /**
     * 从缓存获取Form对象
     * @param formId
     * @return
     */
    public static ModelForm getFromCache(long formId) {
        AbstractCache<ModelForm> cache = CacheManager.getIntance().getCache(ModelForm.ENTITY_NAME);
        if (cache == null) return null;
        return cache.get(formId);
    }

    /**
     * 根据json字符串解析对象
     * @param jsonStr
     * @return
     */
    public static PageInfo parsePageInfo(String jsonStr) {
        if (StringUtils.isEmpty(jsonStr)) return null;
        return JsonUtil.parse(jsonStr, PageInfo.class);
    }

    public static PageInfo parsePageInfo(long formId) {
        ModelForm form = getFromCache(formId);
        if (form == null) return null;
        return parsePageInfo(form.getContent());
    }

    /**
     * 页面传回的content，加工处理成待保存的json串
     * 主要处理内容：
     * 1、model的一些maxlen、默认值等信息，如果和数据库一致，则去掉
     * 2、db中的只保留表、字段和别名
     * @param jsonStr
     * @return
     */
    public static String buildSaveJson(String jsonStr) throws Exception {
        PageInfo pageInfo = parsePageInfo(jsonStr);
        resetDataset(pageInfo);
        for (Map<String, Object> model: pageInfo.model) {
            Map<String, Object> props = (Map<String, Object>)model.get("props");
            buildSaveModelFields(pageInfo, props, true);
            buildSaveModelFields(pageInfo, props, false);
        }
        return encodeSaveJson(pageInfo);
    }

    //处理model的fields和filters
    private static void buildSaveModelFields(PageInfo pageInfo, Map<String, Object> props, boolean isField) {
        String db = (String)props.get("db");
        //没有配置db，配置有误，不处理
        if (StringUtils.isEmpty(db)) throw new SwException("model未配置数据集db，无法解析！");
        PageDataSet pds = pageInfo.findDataSet(db);
        String key = isField ? "fields":"filters";
        List<Map<String, Object>> fields = (List<Map<String, Object>>)props.get(key);
        if (fields == null || fields.isEmpty()) return;
        for (Map<String, Object> fi: fields) {
            Map<String, Object> field = (Map<String, Object>)fi.get("props");
            if (field == null) throw new SwException("model[" + db + "]." + key + "未配置props，无法解析！");

            String fn = (String)field.get("field");
            if (StringUtils.isEmpty(fn)) throw new SwException("model[" + db + "]." + key + "未配置字段名field，无法解析！");
            PageDatasetField pdf = isField ? pds.getField(fn) : pds.getFilter(fn);
            if (pdf == null) throw new SwException("model[" + db + "]." + key + "未找到定义的数据集字段(" + fn + ")，无法解析！");

            //必填字段
            if (pdf.isNotNull() == MapUtil.readBool(field, "required")) {
                field.remove("required");
            }

            if (MapUtil.readInt(field,"maxlength") == SwEnum.DataType.instance.getByValue(pdf.dataType).dataLength) {
                field.remove("maxlength");
            }
        }
    }

    //将db中field按表定义重置一下
    public static void resetDataset(PageInfo pageInfo) {
        for (PageDataSet ds: pageInfo.db) {
            ds.resetFields();
        }
    }

    //构建成需要保存的字符串
    private static String encodeSaveJson(PageInfo pageInfo) throws JsonProcessingException {
        //todo
        ObjectMapper mapper = new ObjectMapper();
        FilterProvider filters = new SimpleFilterProvider().addFilter("datasetField", new SimpleBeanPropertyFilter(){
            private final String includes = "name,label,field,table";
            @Override
            protected boolean include(PropertyWriter writer) {
                return includes.contains(writer.getName());
            }
        }).addFilter("datasetFilter", new SimpleBeanPropertyFilter(){
            private final String includes = ",name,label,field,table,dataType,type,linkDb,linkField,value";
            @Override
            protected boolean include(PropertyWriter writer) {
                return includes.contains(","+ writer.getName());
            }
        });
        return mapper.writer(filters).writeValueAsString(pageInfo);
    }

    /**
     * 根据数据库存储的信息，加工处理成页面要的，与buildsave逆向
     * @param jsonStr
     * @return
     */
    public static String buildReqJson(String jsonStr) {
        PageInfo pageInfo = parsePageInfo(jsonStr);
        resetDataset(pageInfo);
        for (Map<String, Object> model: pageInfo.model) {
            Map<String, Object> props = (Map<String, Object>)model.get("props");
            buildReqModelFields(pageInfo, props, true);
            buildReqModelFields(pageInfo, props, false);
        }
        return JsonUtil.encodeString(pageInfo);
    }

    //处理model的fields和filters
    private static void buildReqModelFields(PageInfo pageInfo, Map<String, Object> props, boolean isField) {
        String db = (String)props.get("db");
        //没有配置db，配置有误，不处理
        if (StringUtils.isEmpty(db)) return;
        PageDataSet pds = pageInfo.findDataSet(db);
        String key = isField ? "fields":"filters";
        List<Map<String, Object>> fields = (List<Map<String, Object>>)props.get(key);
        if (fields == null || fields.isEmpty()) return;
        for (Map<String, Object> fi: fields) {
            Map<String, Object> field = (Map<String, Object>)fi.get("props");
            if (field == null) continue;

            String fn = (String)field.get("field");
            if (StringUtils.isEmpty(fn)) continue;
            PageDatasetField pdf = isField ? pds.getField(fn) : pds.getFilter(fn);
            if (pdf == null) continue;

            //必填字段
            if (!field.containsKey("required")) {
                field.put("required", pdf.isNotNull());
            }

            if (!field.containsKey("maxlength")) {
                field.put("maxlength", SwEnum.DataType.instance.getByValue(pdf.dataType).dataLength);
            }
        }
    }
}
